// minisp

#include <string>
#include <vector>
#include <map>

void warn(const char* err) {
	fputs(err, stderr);
};

enum ValueTypes {
	VT_NULL, VT_INT, VT_SYMBOL, VT_LIST, VT_LAMBDA, VT_NATIVE
};

struct Value;
struct Env;
typedef Value (*NativeFunc)(std::vector<Value>,Env);

struct Value {
	enum ValueTypes    type;
	int                asInt;
	std::string        asString;
	std::vector<Value> asList;
	NativeFunc         asNative;

	bool truthy() {
		switch (type) {
			case VT_NULL: return false;
			case VT_INT:  return asInt != 0;
			case VT_LIST: return asList.size() > 0;
		}
		return true;
	};
	Value operator[] (const int idx) {
		return (asList.size() > idx) ? asList[idx] : Value();
	};
};

struct Env {
	std::map<std::string, Value> dict;
	Env* parent;

	Env(Env* parent) : parent(parent) {};

	void assign(std::string identifier, Value expr) {
		Value* v = resolve(identifier);
		if (v) {
			*v = expr;
		} else {
			dict[identifier] = expr;
		}
	};

	Value lookup(std::string identifier) {
		Value* v = resolve(identifier);
		return v ? *v : Value();
	};

private:
	Value* resolve(std::string identifier) {
		std::map<std::string, Value>::iterator i = dict.find(identifier);
		if (i != dict.end()) {
			return & i->second;
		} else if (parent) {
			return parent->resolve(identifier);
		} else {
			return NULL; // error
		}
	};
};

Value Eval(Value x, Env* env) {
	if (x.type == VT_SYMBOL) {
		return env->lookup(x.asString);

	} else if (x.type != VT_LIST || (x.type == VT_LIST && ! x.asList.size())) {
		// constant literal
		return x;

	} else if (x.asList[0].asString == "quote") {
		x.asList.erase(x.asList.begin());
		return x;

	} else if (x.asList[0].asString == "if") {
		return Eval(Eval(x[1], env).truthy() ? x[2] : x[3], env);

	} else if (x.asList[0].asString == "set!") {
		env->assign(x[1].asString, Eval(x[2], env));
		return Value();

	} else if (x.asList[0].asString == "define") {
		env->dict[x[1].asString] = Eval(x[2], env);
		return Value();

	} else if (x.asList[0].asString == "lambda") {
		// A VT_LAMBDA is stored as a list with two entries. The first entry
		//    must be a list containing only symbols.
		Value ret;
		ret.type = VT_LAMBDA;
		ret.asList.push_back(x[1]);
		ret.asList.push_back(x[2]);
		return ret;

	} else if (x.asList[0].asString == "begin") {
		Env block_scope = Env(env);
		Value ret;
		for (int i = 1, e = x.asList.size(); i < e; i++) {
			ret = Eval(x.asList[i], &block_scope);
		}
		return ret;

	} else {
		// function call
		
		Value proc = Eval(x[0], env);
		std::vector<Value> args;
		for (int i = 1, e = x.asList.size(); i < e; i++) {
			args.push_back( Eval(x[i], env) );
		}

		if (proc.type == VT_NATIVE) {
			return proc.asNative(args, env);

		} else if (proc.type == VT_LAMBDA) {
			Env lambda_env(NULL); // not a closure
			for (int i = 0, e = proc[1].asList.size(); i < e; i++) {
				lambda_env.dict[proc[1].asList[i].asString] = (args.size() > i) ? args[i] : Value();
			}
			return Eval(proc[2], &lambda_env);

		} else {
			return proc;
		}

	}
};

bool isWhitespace(char c) {
	return c == ' ' || c == '\n' || c == '\r' || c == '\t';
}

Value atom(std::string token) {
	Value ret;

	if (! token.size()) {
		return ret;	

	} else if (token[0] >= '0' && token[0] <= '9') {
		ret.type = VT_INT;
		ret.asInt = atoi(token.c_str());

	} else {
		ret.type = VT_SYMBOL;
		ret.asString = token;

	}
	return ret;
}

Value parseSymbol(std::string *src, int *pos) {
	std::string word;
	for(;;) {
		if (*pos >= src->length()) {
			return atom(word);
		}

		char c = src->at(*pos);
		if (isWhitespace(c) || c == '(' || c == ')') {
			return atom(word);
		}

		word.push_back(c);

		(*pos)++;
	}
}

void parseWhitespace(std::string *src, int *pos) {
	int len = src->length();
	while( *pos < len && isWhitespace(src->at(*pos))) (*pos)++;
}

Value parseList(std::string *src, int *pos) {
	Value ret;
	ret.type = VT_LIST;

	for(;;) {

		parseWhitespace(src, pos);

		if (*pos >= src->length()) {
			return ret;
		}
		
		char c = src->at(*pos);
		if (c == '(') {
			(*pos)++;
			ret.asList.push_back(parseList(src, pos));

		} else if (c == ')') {
			(*pos)++;
			return ret;

		} else {
			ret.asList.push_back(parseSymbol(src, pos));

		}
	}
};

Value parse(std::string src) {
	int pos = 0;
	Value ret = parseList(&src, &pos);
	parseWhitespace(&src, &pos);

	if (pos < src.length()) {
		warn("Didn't completely parse input");
	}
	return ret;
};

// ############################################################################

Value myNativeAdd(std::vector<Value> args, Env e) {
	Value ret;
	ret.type = VT_INT;
	ret.asInt = 0;

	for (int i = 0, e = args.size(); i < e; i++) {
		if (args[i].type == VT_INT) ret.asInt += args[i].asInt;
	}

	return ret;
}

int main(int argc, char** argv) {

	Value c = parse("(add (add 2 3) 4)");

	Value nativeAdd;
	nativeAdd.type = VT_NATIVE;
	nativeAdd.asNative = myNativeAdd;

	Env e(NULL);
	e.assign("add", nativeAdd);

	Value ret = Eval(c, &e);
}